

from typing import Any, Callable, Coroutine, Literal, Optional

from eth_utils.toolz import assoc, merge
from web3.types import Formatters, FormattersDict, RPCEndpoint, RPCResponse

from dank_mids.types import AsyncMiddleware
from web3 import Web3

FORMATTER_DEFAULTS: FormattersDict = {
    "request_formatters": {},
    "result_formatters": {},
    "error_formatters": {},
}

async def async_construct_formatting_middleware(
    request_formatters: Optional[Formatters] = None,
    result_formatters: Optional[Formatters] = None,
    error_formatters: Optional[Formatters] = None,
) -> AsyncMiddleware:
    async def ignore_web3_in_standard_formatters(
        _w3: "Web3",
        _method: RPCEndpoint,
    ) -> FormattersDict:
        return dict(
            request_formatters=request_formatters or {},
            result_formatters=result_formatters or {},
            error_formatters=error_formatters or {},
        )

    return await async_construct_web3_formatting_middleware(  # type: ignore [return-value]
        ignore_web3_in_standard_formatters
    )

async def async_construct_web3_formatting_middleware(
    async_web3_formatters_builder: Callable[
        ["Web3", RPCEndpoint], Coroutine[Any, Any, FormattersDict]
    ]
) -> Callable[
    [Callable[[RPCEndpoint, Any], Any], "Web3"], Coroutine[Any, Any, AsyncMiddleware]
]:
    async def formatter_middleware(
        make_request: Callable[[RPCEndpoint, Any], Any],
        async_w3: "Web3",
    ) -> AsyncMiddleware:
        async def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
            formatters = merge(
                FORMATTER_DEFAULTS,
                await async_web3_formatters_builder(async_w3, method),
            )
            request_formatters = formatters.pop("request_formatters")

            response = await make_request(
                method,
                request_formatters[method](params) if method in request_formatters else params,
            )

            return _apply_response_formatters(
                method=method, response=response, **formatters
            )

        return middleware

    return formatter_middleware

def _apply_response_formatters(
    method: RPCEndpoint,
    response: RPCResponse,
    result_formatters: Formatters,
    error_formatters: Formatters,
) -> RPCResponse:
    def _format_response(
        response_type: Literal["result", "error"],
        method_response_formatter: Callable[..., Any],
    ) -> RPCResponse:
        appropriate_response = response[response_type]
        return assoc(
            response, response_type, method_response_formatter(appropriate_response)
        )

    if "result" in response and method in result_formatters:
        return _format_response("result", result_formatters[method])
    elif "error" in response and method in error_formatters:
        return _format_response("error", error_formatters[method])
    else:
        return response



import threading

class UIDGenerator:
    def __init__(self) -> None:
        self._value: int = -1
        self.lock = threading.Lock()
    
    @property
    def latest(self) -> int:
        """ Returns the latest unique id. """
        return self._value
    
    @property
    def next(self) -> int:
        """ Returns the next unique id. """
        with self.lock:
            new: int = self._value + 1
            self._value = new
            return new




import asyncio
import threading
from collections import defaultdict
from time import time
from typing import Any, DefaultDict, List

from eth_utils import to_checksum_address
from multicall.constants import MULTICALL2_ADDRESSES, MULTICALL_ADDRESSES
from multicall.multicall import NotSoBrightBatcher
from web3 import Web3
from web3.providers import HTTPProvider
from web3.providers.async_base import AsyncBaseProvider
from web3.types import RPCEndpoint, RPCResponse

from dank_mids._config import LOOP_INTERVAL
from dank_mids._demo_mode import demo_logger
from dank_mids.helpers import _session
from dank_mids.loggers import main_logger, sort_lazy_logger
from dank_mids.requests import JSONRPCBatch, Multicall, RPCRequest, eth_call
from dank_mids.types import BlockId, ChainId
from dank_mids.uid import UIDGenerator
from dank_mids.worker import DankWorker

BYPASS_METHODS = "eth_getLogs", "trace_", "debug_"

instances: DefaultDict[ChainId, List["DankMiddlewareController"]] = defaultdict(list)

def _sync_w3_from_async(w3: Web3) -> Web3:
    assert w3.eth.is_async and isinstance(w3.provider, AsyncBaseProvider), "Dank Middleware can only be applied to an asycnhronous Web3 instance."
    sync_w3: Web3 = Web3(provider = HTTPProvider(w3.provider.endpoint_uri, {'timeout': 600}, session=_session.get_session()))
    # We can't pickle middlewares to send to process executor.
    # The call has already passed thru all middlewares on the user's Web3 instance.
    sync_w3.middleware_onion.clear()
    sync_w3.provider.middlewares = tuple()
    return sync_w3


class DankMiddlewareController:
    def __init__(self, w3: Web3) -> None:
        main_logger.info('Dank Middleware initializing... Strap on your rocket boots...')
        self.w3: Web3 = w3
        self.sync_w3 = _sync_w3_from_async(w3)
        self.chain_id = self.sync_w3.eth.chain_id
        multicall = MULTICALL_ADDRESSES.get(self.chain_id)
        multicall2 = MULTICALL2_ADDRESSES.get(self.chain_id)
        if multicall2 is None:
            raise NotImplementedError("Dank Mids currently does not support this network.\nTo add support, you just need to submit a PR adding the appropriate multicall contract addresses to this file:\nhttps://github.com/banteg/multicall.py/blob/master/multicall/constants.py")
        self.multicall2 = to_checksum_address(multicall2)
        self.no_multicall = {self.multicall2} if multicall is None else {self.multicall2, to_checksum_address(multicall)}
        self.call_uid = UIDGenerator()
        # NOTE: We need some internal things set up before we can start the worker thread.
        self._futs: List[asyncio.Future] = []
        self._daemon_running = False
        self._wakeup_loop = asyncio.get_event_loop()._write_to_self
        self.worker = DankWorker(self)
        self.pending_rpc_calls = JSONRPCBatch(self.worker)
        self.pending_eth_calls: DefaultDict[BlockId, Multicall] = defaultdict(lambda: Multicall(self.worker))
        self._pools_closed_lock = threading.Lock()
        self._first = None
        self._checkpoint: float = time()
        self._instance: int = sum(len(_instances) for _instances in instances.values())
        instances[self.chain_id].append(self)  # type: ignore
    
    def __repr__(self) -> str:
        return f"<DankMiddlewareController instance={self._instance} chain={self.chain_id} endpoint={self.endpoint}>"

    async def __call__(self, method: RPCEndpoint, params: Any) -> RPCResponse:
        self._start_exception_daemon_if_stopped()
        if method == "eth_call":
            return await eth_call(self, params)
        return await RPCRequest(self, method, params)
    
    @property
    def endpoint(self) -> str:
        return self.w3.provider.endpoint_uri  # type: ignore
    
    @property
    def batcher(self) -> NotSoBrightBatcher:
        return self.worker.batcher
    
    async def execute_multicall(self) -> None:
        self.checkpoint = time()
        i = 0
        while self._pools_closed_lock.locked():
            if i // 500 == int(i // 500):
                main_logger.debug('lock is locked')
            i += 1
            await asyncio.sleep(.1)
        
        # NOTE we put this here to prevent a double locking issue
        empty = JSONRPCBatch(self.worker)
        with self._pools_closed_lock:
            eth_calls = dict(self.pending_eth_calls)
            self.pending_eth_calls.clear()
            rpc_calls = self.pending_rpc_calls
            self.pending_rpc_calls = empty
        demo_logger.info(f'executing multicall (current cid: {self.call_uid.latest})')  # type: ignore
        await self.worker.execute_batch(eth_calls, rpc_calls)

    @sort_lazy_logger
    def should_batch(self, method: RPCEndpoint, params: Any) -> bool:
        """ Determines whether or not a call should be passed to the DankMiddlewareController. """
        if method == "eth_call" and params[0]["to"] == self.multicall2:
            # These most likely come from dank mids internals if you're using dank mids.
            return False
        if any(bypass in method for bypass in BYPASS_METHODS):
            main_logger.debug(f"bypassed, method is {method}")
            return False
        return True

    @property
    def loop_is_ready(self) -> bool:
        return time() - self._checkpoint > LOOP_INTERVAL
    
    @property
    def queue_is_full(self) -> bool:
        return sum(len(v) for v in self.pending_eth_calls.values()) >= self.batcher.step * 25

    def reduce_batch_size(self, num_calls: int) -> None:
        new_step = round(num_calls * 0.99) if num_calls >= 100 else num_calls - 1
        # NOTE: We need this check because one of the other multicalls in a batch might have already reduced `self.batcher.step`
        if new_step < self.batcher.step:
            old_step = self.batcher.step
            self.batcher.step = new_step
            main_logger.warning(f'Multicall batch size reduced from {old_step} to {new_step}. The failed batch had {num_calls} calls.')
    
    def _start_exception_daemon_if_stopped(self) -> None:
        if self._daemon_running:
            return
        
        def done_callback(fut: asyncio.Future) -> None:
            """Notifies the controller in the event of daemon shutdown or failure."""
            self._daemon_running = False
            if not fut.exception():
                try:
                    self._futs.remove(fut)
                except ValueError as e:
                    if str(e) != "list.remove(x): x not in list":
                        raise
                    
        fut = asyncio.ensure_future(self._exception_daemon())
        fut.add_done_callback(done_callback)
        self._futs.append(fut)
            
    async def _exception_daemon(self) -> None:
        if self._daemon_running:
            # Just in case
            return
        main_logger.debug('starting exception daemon.')
        self._daemon_running = True
        while self._futs:
            futs = self._futs[:]
            main_logger.debug(f'{sum(1 for fut in self._futs if fut.done())} futs are done and ready to pop')
            if any(fut.done() and fut.exception() for fut in futs):
                main_logger.critical(f"DankMids futures have err'd: {self._futs}")
                [main_logger.critical(f"DankMidsError: {e}") for fut in futs if fut.done() and (e := fut.exception())]
            else:
                main_logger.debug(self._futs)
            for fut in futs:
                if fut.done() and not fut.exception():
                    try:
                        self._futs.remove(fut)
                    except ValueError as e:
                        if str(e) != "list.remove(x): x not in list":
                            raise
                    
            await asyncio.sleep(5)
        main_logger.debug('exiting exception daemon.')





from functools import lru_cache, partial
from typing import Any, Callable, List, Tuple, Union

from brownie.convert.normalize import (ABIType, TupleType, _check_array,
                                       _format_single, _get_abi_types)


@lru_cache(maxsize=None)
def get_abi_formatter(abi_inputs: List[dict]) -> Callable[[tuple], tuple]:
    formatters = [_get_formatter_fn(arg_type) for arg_type in _get_abi_types(abi_inputs)]
    def abi_formatter(args):
        return [formatter(arg) for arg, formatter in zip(args, formatters)]
    return abi_formatter

_formatters = {}
def _get_formatter_fn(_type: ABIType) -> Callable[[Any], Any]:
    """This eliminates some of the standard web3.py formatter logic without sacrificing uality."""
    type_str = _type.to_type_str()
    if type_str not in _formatters:
        if _type.is_array:
            _formatters[type_str] = _get_array_formatter(_type)
        elif isinstance(_type, TupleType):
            _formatters[type_str] = _get_tuple_formatter(_type)
        else:
            _formatters[type_str] = partial(_format_single, _type.to_type_str())
    return _formatters[type_str]

_array_formatters = {}
def _get_array_formatter(typ: ABIType) -> Callable[[Any], Any]:
    item_type = typ.item_type
    type_str = typ.to_type_str()
    if type_str not in _array_formatters:
        check_array = partial(_check_array, length=typ.arrlist[-1][0] if len(typ.arrlist[-1]) else None)
        
        if item_type.is_array:
            array_formatter = _get_array_formatter(item_type)
            def _format_array(values: Union[List, Tuple]) -> List:
                check_array(values)
                return [array_formatter(i) for i in values]
            
        elif isinstance(item_type, TupleType):
            tuple_formatter = _get_tuple_formatter(item_type)
            def _format_array(values: Union[List, Tuple]) -> List:
                check_array(values)
                return [tuple_formatter(i) for i in values]
        else:
            def _format_array(values: Union[List, Tuple]) -> List:
                check_array(values)
                return [_format_single(item_type.to_type_str(), i) for i in values]
        _array_formatters[type_str] = _format_array
    return _array_formatters[type_str]

_tuple_formatters = {}
def _get_tuple_formatter(typ: TupleType):
    if not isinstance(typ, TupleType):
        raise TypeError(typ) 
    type_str = typ.to_type_str()
    if type_str not in _tuple_formatters:
        formatters = [_get_formatter_fn(_) for _ in typ.components]
        check_array = partial(_check_array, length=len(formatters))
        def _format_tuple(values: Union[List, Tuple]) -> List:
            result = []
            check_array(values)
            for formatter, value in zip(formatters, values):
                try:
                    result.append(formatter(value))
                except Exception as e:
                    raise type(e)(f"'{value}' - {e}") from None
            return result
        _tuple_formatters[type_str] = _format_tuple
    return _tuple_formatters[type_str]




from brownie import Contract, web3
from dank_mids.brownie_patch import patch_contract
from dank_mids.brownie_patch.call import _patch_call
from dank_mids import setup_dank_w3_from_sync
from multicall.utils import await_awaitable

from tests.fixtures import dank_w3


def test_patch_call():
    # must use from_explorer for gh testing workflow
    weth = Contract.from_explorer('0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2')
    _patch_call(weth.totalSupply, dank_w3)
    assert hasattr(weth.totalSupply, 'coroutine')
    assert await_awaitable(weth.totalSupply.coroutine(block_identifier=13_000_000)) == 6620041514474872981393155

def test_patch_contract():
    # ContractCall
    # must use from_explorer for gh testing workflow
    weth = patch_contract(Contract.from_explorer('0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2'), dank_w3)
    assert hasattr(weth.totalSupply, 'coroutine')
    assert await_awaitable(weth.totalSupply.coroutine(block_identifier=13_000_000)) == 6620041514474872981393155

    # ContractTx
    # must use from_explorer for gh testing workflow
    uni_v3_quoter = patch_contract(Contract.from_explorer('0xb27308f9F90D607463bb33eA1BeBb41C27CE5AB6'), dank_w3)
    assert hasattr(uni_v3_quoter.quoteExactInput, 'coroutine')
    assert await_awaitable(
        uni_v3_quoter.quoteExactInput.coroutine(b"\xc0*\xaa9\xb2#\xfe\x8d\n\x0e\\O'\xea\xd9\x08<ul\xc2\x00\x01\xf4\xa0\xb8i\x91\xc6!\x8b6\xc1\xd1\x9dJ.\x9e\xb0\xce6\x06\xebH", 1e18, block_identifier=13_000_000)
    ) == 3169438072

def test_call_setup_twice_on_same_web3():
    w3_a = setup_dank_w3_from_sync(web3)
    w3_a.test = True
    w3_b = setup_dank_w3_from_sync(web3)
    assert hasattr(w3_b, 'test')
    


import asyncio
from typing import Union
from web3.types import RPCEndpoint
from dank_mids._config import semaphore_envs
from threading import current_thread
from typing import Dict


class ThreadsafeSemaphore:
    """
    While its a bit weird to run multiple event loops, sometimes either you or a lib you're using must do so. 
    When in use in threaded applications, the semaphore will not work as intended but at least your program will function.
    You may need to reduce the semaphore value for multi-threaded applications.
    """

    def __init__(self, value: int) -> None:
        self.default_value = value
        self.semaphores: Dict[int, asyncio.Semaphore] = {}
    
    @property
    def semaphore(self) -> asyncio.Semaphore:
        tid = current_thread()
        if tid not in self.semaphores:
            self.semaphores[tid] = asyncio.Semaphore(self.default_value)  # type: ignore [index]
        return self.semaphores[tid]  # type: ignore [index]
    
    async def __aenter__(self):
        await self.semaphore.acquire()
    
    async def __aexit__(self, *args):
        self.semaphore.release()


class _DummySemaphore:
    async def __aenter__(self):
        ...
    async def __aexit__(self, *args):
        ...

class Semaphores:
    def __init__(self) -> None:
        self.method_semaphores = {key: ThreadsafeSemaphore(value) for key, value in semaphore_envs.items()}
        self.keys = self.method_semaphores.keys()
        self.dummy = _DummySemaphore()
    
    def __getitem__(self, method: RPCEndpoint) -> Union[ThreadsafeSemaphore, _DummySemaphore]:
        for key in self.keys:
            if key in method:
                return self.method_semaphores[key]
        return self.dummy

method_semaphores = Semaphores()




import os

from aiohttp import ClientTimeout

LOOP_INTERVAL = float(os.environ.get("DANKMIDSLOOPINTERVAL", 0.01))

# Max number of rpc calls to include in one batch call
MAX_JSONRPC_BATCH_SIZE = int(os.environ.get("MAX_JSONRPC_BATCH_SIZE", 500))


# Enable Demo Mode?
DEMO_MODE = bool(os.environ.get("DANKMIDS_DEMO_MODE", os.environ.get("DANK_MIDS_DEMO_MODE")))

# Are you calling a ganache fork? Can't use state override code
GANACHE_FORK = bool(os.environ.get("DANKMIDS_GANACHE_FORK", os.environ.get("DANK_MIDS_GANACHE_FORK")))

# With default AsyncBaseProvider settings, some dense calls will fail
#  due to aiohttp.TimeoutError where they would otherwise succeed.
# This is not due to the response time from the node, but the event loop timing.
# We set the default to 5 minutes but if you're doing serious work
#  you may want to increase it further.
AIOHTTP_TIMEOUT = ClientTimeout(int(os.environ.get("AIOHTTP_TIMEOUT", 300)))


# Method-specific Semaphores
semaphore_envs = {
    "eth_getBlock": int(os.environ.get("ETH_GETBLOCK_SEMAPHORE", 50)), # [eth_getBlockByNumber, eth_getBlockByHash]
    "eth_getLogs": int(os.environ.get("ETH_GETLOGS_SEMAPHORE", 16)),
    "eth_getTransaction": int(os.environ.get("ETH_GETTRANSACTION_SEMAPHORE", 100)),
}

# Brownie call Semaphore
#   Used because I experienced some OOM errs due to web3 formatters when I was batching an absurd number of brownie calls.
#   We need a separate semaphore here because the method-specific semaphores are too late in the code to prevent this OOM issue.
BROWNIE_CALL_SEMAPHORE_VAL = int(os.environ.get("DANK_MIDS_BROWNIE_CALL_SEMAPHORE", 100_000))




import abc
import asyncio
import threading
from collections import defaultdict
from functools import partial
from time import time
from typing import (TYPE_CHECKING, Any, DefaultDict, Dict, Generator, Generic,
                    Iterable, Iterator, List, Optional, Tuple, TypeVar, Union)

import aiohttp
import eth_retry
import requests
from aiohttp import RequestInfo
from eth_abi import decode_single, encode_single
from eth_typing import ChecksumAddress
from eth_utils import function_signature_to_4byte_selector
from hexbytes import HexBytes
from hexbytes._utils import to_bytes
from multicall.utils import run_in_subprocess
from web3 import Web3
from web3.datastructures import AttributeDict
from web3.types import RPCEndpoint, RPCError, RPCResponse

from dank_mids._config import (AIOHTTP_TIMEOUT, DEMO_MODE,
                               MAX_JSONRPC_BATCH_SIZE)
from dank_mids._demo_mode import demo_logger
from dank_mids.constants import BAD_HEXES, OVERRIDE_CODE
from dank_mids.helpers import _session, await_all
from dank_mids.loggers import main_logger
from dank_mids.types import BatchId, BlockId, JsonrpcParams, RpcCallJson, T

if TYPE_CHECKING:
    from dank_mids.controller import DankMiddlewareController
    from dank_mids.worker import DankWorker

RETRY_ERRS = ["connection reset by peer","request entity too large","server disconnected","execution aborted (timeout = 5s)"]

class ResponseNotReady(Exception):
    pass

def _call_failed(data: Optional[Union[bytes, Exception]]) -> bool:
    """ Returns True if `data` indicates a failed response, False otherwise. """
    if data is None:
        return True
    # If we got a known "bad" response from the multicall, that is also a failure.
    # Most likely the target contract does not support multicall2.
    elif (isinstance(data, bytes) and HexBytes(data).hex() in BAD_HEXES):
        return True
    return False

def _reattempt_call_and_return_exception(target: ChecksumAddress, calldata: bytes, block: BlockId, w3: Web3) -> Union[bytes, Exception]:
    """ NOTE: This runs synchronously in a subprocess in order to bypass Dank Middleware without blocking the event loop. """
    try:
        return w3.eth.call({"to": target, "data": calldata}, block)
    except Exception as e:
        return e

def _err_response(e: Exception) -> RPCError:
    """ Extract an error message from `e` to use in a spoof rpc response. """
    if isinstance(e.args[0], (str, RequestInfo)):
        if isinstance(e, AttributeError):
            raise e
        err_msg = f"DankMidsError: {e.__class__.__name__}: {e.args}"
    elif isinstance(e.args[0], AttributeError):
        raise e.args[0]
    elif isinstance(e.args[0], Exception):
        err_msg = f"DankMidsError: {e.args[0].__class__.__name__}: {e.args[0].args}"
    elif not hasattr(e.args[0], '__contains__'):
        if isinstance(e, AttributeError):
            raise e
        err_msg = f"DankMidsError: {e.__class__.__name__}: {e.args}"
    elif "message" in e.args[0]:
        err_msg = e.args[0]["message"]
    elif "error" in e.args[0] and hasattr(e.args[0]["error"], '__contains__') and "message" in e.args[0]["error"]:
        err_msg = e.args[0]["error"]["message"]
    else:
        raise e
    return {'code': -32000, 'message': err_msg, 'data': ''}


_Response = TypeVar("_Response", RPCResponse, List[RPCResponse])

class _RequestMeta(Generic[_Response], metaclass=abc.ABCMeta):
    batch: "_Batch"
    controller: "DankMiddlewareController"
    def __init__(self) -> None:
        self.uid = self.controller.call_uid.next
        self._started = False
        self._response = None
        self._done = asyncio.Event()
    
    def __await__(self) -> Generator[Any, None, Optional[_Response]]:
        return self.get_response().__await__()
    
    @abc.abstractmethod
    def __len__(self) -> int:
        pass

    @abc.abstractmethod
    async def get_response(self) -> Optional[_Response]:
        pass

    def set_done(self, wakeup: bool = True) -> None:
        self._done.set()
        if wakeup:
            # NOTE: We are in a separate thread from the asyncio.Event's event loop.
            #       In this case we must wake up the event loop manually.
            self.controller._wakeup_loop()

### Single requests:

single_semaphore = asyncio.Semaphore(1)
    
class RPCRequest(_RequestMeta[RPCResponse]):
    def __init__(self, controller: "DankMiddlewareController", method: RPCEndpoint, params: Any):
        self.controller = controller
        self.controller.checkpoint = time()
        self.method = method
        self.params = params
        super().__init__()
        
        if isinstance(self, eth_call) and self.multicall_compatible:
            self.controller.pending_eth_calls[self.block].append(self)
        else:
            self.controller.pending_rpc_calls.append(self)
        demo_logger.info(f'added to queue (cid: {self.uid})')  # type: ignore
    
    def __eq__(self, __o: object) -> bool:
        return self.uid == __o.uid if isinstance(__o, self.__class__) else False 
    
    def __hash__(self) -> int:
        return self.uid
    
    def __len__(self) -> int:
        return 1
    
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} uid={self.uid} method={self.method}>"

    @property
    def rpc_data(self) -> RpcCallJson:
        return {'jsonrpc': '2.0', 'id': self.uid, 'method': self.method, 'params': self.params}
            
    async def get_response(self) -> RPCResponse:
        if not self._started:
            if self.controller._first is None:
                self.controller._first = self
                await asyncio.sleep(0)
            if self.controller._first is self:
                while not self.controller.loop_is_ready:
                    await asyncio.sleep(0)
                await self.controller.execute_multicall()
                self.controller._first = None
        await self._done.wait()
        return await self.spoof_response(self._response)
    
    async def set_response(self, data: Union[str, AttributeDict, Exception], wakeup_main_loop: bool = True) -> None:
        self._response = data
    
    async def spoof_response(self, data: Union[bytes, str, AttributeDict, Exception]) -> RPCResponse:
        spoof = {"id": self.uid, "jsonrpc": "dank_mids"}
        if isinstance(data, bytes):
            spoof["result"] = data.hex()  # type: ignore
        elif isinstance(data, (str, AttributeDict)):
            spoof["result"] = data
        elif isinstance(data, Exception):
            spoof["error"] = _err_response(data)
        else:
            raise NotImplementedError(data.__class__.__name__)
        if isinstance(self, eth_call):
            main_logger.debug(f"method: eth_call  address: {self.target}  spoof: {spoof}")
        else:
            main_logger.debug(f"method: {self.method}  spoof: {spoof}")
        return spoof
            

class eth_call(RPCRequest):
    def __init__(self, controller: "DankMiddlewareController", params: Any) -> None:
        """ Adds a call to the DankMiddlewareContoller's `pending_eth_calls`. """
        super().__init__(controller, "eth_call", params)  # type: ignore
    
    @property
    def block(self) -> BlockId:
        return self.params[1]
    
    @property
    def calldata(self) -> HexBytes:
        return HexBytes(self.params[0]['data'])
    
    @property
    def multicall_compatible(self) -> bool:
        return self.target not in self.controller.no_multicall

    @property
    def target(self) -> str:
        return self.params[0]["to"]

    async def spoof_response(self, data: Union[bytes, Exception]) -> RPCResponse:  # type: ignore
        """ Sets and returns a spoof rpc response for this BatchedCall instance using data provided by the worker. """
        # NOTE: If multicall failed, make sync call to get either:
        # - revert details
        # - successful response
        if _call_failed(data):
            data = await self.sync_call()
        return await super().spoof_response(data)

    async def sync_call(self) -> Union[bytes, Exception]:
        """ Used to bypass DankMiddlewareController. """
        data = await self.controller.worker.run_in_executor(
            _reattempt_call_and_return_exception, self.target, self.calldata, self.block, self.controller.sync_w3, loop=asyncio.get_event_loop()
        )
        # If we were able to get a usable response from single call, add contract to `do_not_batch`.
        if not isinstance(data, Exception):
            self.controller.no_multicall.add(self.target)  # type: ignore
        return data


### Batch requests:

_Request = TypeVar("_Request")

class _Batch(_RequestMeta[List[RPCResponse]], Iterable[_Request]):
    calls: List[_Request]

    def __init__(self, worker: "DankWorker", calls: Iterable[_Request]):
        self.worker: DankWorker = worker
        self.controller: "DankMiddlewareController" = worker.controller
        self.calls = list(calls)  # type: ignore
        for call in self.calls:
            call.batch = self
        super().__init__()
        self._len = len(self.calls)
        self._len_lock = threading.Lock()
        self._fut = None
        self._done = asyncio.Event()
    
    def set_done(self, wakeup: bool = True) -> None:
        self._done.set()
        for call in self.calls:
            call.set_done(wakeup=wakeup)
        # NOTE: We are in a separate thread from the asyncio.Event's event loop.
        #       In this case we must wake up the event loop in the main thread.
        self.controller._wakeup_loop()
    
    def __bool__(self) -> bool:
        return bool(self.calls)
    
    def __getitem__(self, ix: int) -> _Request:
        return self.calls[ix]

    def __iter__(self) -> Iterator[_Request]:
        return iter(self.calls)
    
    def __len__(self) -> int:
        with self._len_lock:
            return self._len
    
    async def coroutine(self) -> None:
        return await self
    
    def append(self, call: _Request, skip_check: bool = False) -> None:
        self.calls.append(call)
        call.batch = self
        with self._len_lock:
            self._len += 1
        if not skip_check and self.is_full:
            self.ensure_future()
    
    def extend(self, calls: Iterable[_Request], skip_check: bool = False) -> None:
        self.calls.extend(calls)
        for call in calls:
            call.batch = self
        with self._len_lock:
            self._len += len(calls)
        if not skip_check and self.is_full:
            self.ensure_future()
    
    def ensure_future(self) -> asyncio.Future:
        if self._fut is not None:
            return self._fut
        with self.controller._pools_closed_lock:
            self._ensure_future()
        return self._fut
    
    def _ensure_future(self) -> asyncio.Future:
        """Not threadsafe"""
        for call in self.calls:
            call._started = True
            if isinstance(call, Multicall):
                for _call in call:
                    _call._started = True
        self._fut = self.worker.run_coro_threadsafe(self.coroutine())
        self.controller._futs.append(self._fut)
        self._post_ensure_future_cleanup()
    
    @property
    def halfpoint(self) -> int:
        return len(self) // 2
    
    def bisect(self) -> Generator[List[_Request], None, None]:
        """Destroys this reuest."""
        yield self.chunk0
        yield self.chunk1
        
    @property
    def chunk0(self) -> List[_Request]:
        return self.calls[:self.halfpoint]
    
    @property
    def chunk1(self) -> List[_Request]:
        return self.calls[self.halfpoint:]
    
    def should_retry(self, e: Exception) -> bool:
        if "out of gas" in f"{e}":
            # TODO Remember which contracts/calls are gas guzzlers
            main_logger.debug('out of gas. cut in half, trying again')
        elif any(err in f"{e}".lower() for err in RETRY_ERRS):
            # TODO: use these exceptions to optimize for the user's node
            main_logger.debug('Dank too loud. Bisecting batch and retrying.')
        elif "error processing call Revert" not in f"{e}":
            main_logger.warning(f"unexpected {e.__class__.__name__}: {e}")
        return len(self) > 1


class Multicall(_Batch[eth_call]):
    """ Runs in worker thread. One-time use."""
    method = "eth_call"
    fourbyte = function_signature_to_4byte_selector("tryBlockAndAggregate(bool,(address,bytes)[])")
    encode_args = partial(encode_single, "(bool,(address,bytes)[])")
    decode_response = partial(run_in_subprocess, partial(decode_single, "(uint256,uint256,(bool,bytes)[])"))

    def __init__(
        self,
        worker: "DankWorker",
        calls: List[eth_call] = [],
        bid: Optional[BatchId] = None
    ) -> None:
        super().__init__(worker, calls)
        self.bid = bid or self.worker.multicall_uid.next
    
    def __repr__(self) -> str:
        return f"<Multicall mid={self.bid} block={self.block} len={len(self)}>"
    
    @property
    def block(self) -> BlockId:
        return self.calls[0].block
    
    @property
    def calldata(self) -> str:
        return (self.fourbyte + Multicall.encode_args([False, [[call.target, call.calldata] for call in self.calls]])).hex()
    
    @property
    def target(self) -> ChecksumAddress:
        return self.worker.target
    
    @property
    def params(self) -> JsonrpcParams:
        params = [{'to': self.target, 'data': f'0x{self.calldata}'}, self.block]
        if self.worker.state_override_supported:
            params.append({self.target: {'code': OVERRIDE_CODE}})
        return params
    
    @property
    def rpc_data(self) -> RpcCallJson:
        return {'jsonrpc': '2.0', 'id': self.uid, 'method': self.method, 'params': self.params}
    
    @property
    def is_full(self) -> bool:
        return len(self) >= self.controller.batcher.step
            
    async def get_response(self) -> None:
        rid = self.worker.request_uid.next
        demo_logger.info(f'request {rid} for multicall {self.bid} starting')  # type: ignore
        try:
            await self.set_response(await self.worker(*self.params))
        except Exception as e:
            await (self.bisect_and_retry() if self.should_retry(e) else self.set_response(e))  # type: ignore [misc]
        demo_logger.info(f'request {rid} for multicall {self.bid} complete')  # type: ignore
    
    def should_retry(self, e: Exception) -> bool:
        if any(err in f"{e}".lower() for err in RETRY_ERRS):
            main_logger.debug('dank too loud, trying again')
            self.controller.reduce_batch_size(len(self))
            return True
        elif "No state available for block" in f"{e}":  # NOTE: While it might look weird, f-string is faster than `str(e)`.
            e.args[0]["dankmids_note"] = "You're not using an archive node, and you need one for the application you are attempting to run."
            return False
        elif super().should_retry(e):
            return True
        return len(self) > 1
    
    async def set_response(self, data: Union[bytes, str, Exception], wakeup_main_loop: bool = True) -> None:
        """
        If called from `self`, `response` will be bytes type.
        if called from a JSONRPCBatch, `response` will be str type.
        """ 
        if isinstance(data, Exception):
            await await_all(call.set_response(data, wakeup_main_loop=False) for call in self.calls)
        else:
            decoded: List[Tuple[bool, bytes]]
            _, _, decoded = await Multicall.decode_response(to_bytes(data))
            await await_all(call.set_response(data, wakeup_main_loop=False) for call, (_, data) in zip(self.calls, decoded))
        if wakeup_main_loop:
            self.set_done()
    
    async def bisect_and_retry(self) -> List[RPCResponse]:
        await await_all(Multicall(self.worker, chunk, f"{self.bid}_{i}") for i, chunk in enumerate(self.bisect()))
    
    def _post_ensure_future_cleanup(self) -> None:
        self.controller.pending_eth_calls.pop(self.block)


session = _session.get_session()

def post_sync(endpoint, data) -> Union[bytes, Tuple[str, Exception]]:
    response = session.post(endpoint, json=data)
    try:
        response.raise_for_status()
        return response.json()
    except requests.HTTPError as e:
        return e.args[0], e
    except Exception as e:
        # This really shouldn't be running but just in case
        return response.reason, e
    
class JSONRPCBatch(_Batch[Union[Multicall, RPCRequest]]):
    def __init__(self, worker: "DankWorker", calls: List[Union[Multicall, RPCRequest]] = [], jid: Optional[BatchId] = None) -> None:
        super().__init__(worker, calls)
        self.jid = jid or self.worker.jsonrpc_batch_uid.next
        self._fut: Optional[asyncio.Future] = None
    
    @property
    def data(self) -> List[RpcCallJson]:
        return [call.rpc_data for call in self.calls]
    
    @property
    def is_multicalls_only(self) -> bool:
        return all(isinstance(call, Multicall) for call in self.calls)

    @property
    def is_single_multicall(self) -> bool:
        return len(self) == 1 and self.is_multicalls_only

    @property
    def method_counts(self) -> Dict[RPCEndpoint, int]:
        counts: DefaultDict[RPCEndpoint, int] = defaultdict(int)
        for call in self.calls:
            counts[call.method] += len(call)  # type: ignore
        return dict(counts)
    
    @property
    def total_calls(self) -> int:
        return sum(len(call) for call in self.calls)
    
    @property
    def is_full(self) -> bool:
        return len(self) >= MAX_JSONRPC_BATCH_SIZE
    
    def extend(self, calls: "JSONRPCBatch", skip_check: bool=False) -> None:
        if calls.is_single_multicall:
            return self.append(calls, skip_check=skip_check)
        super().extend(calls, skip_check=skip_check)
    
    async def get_response(self) -> None:
        """ Runs in worker thread. """
        rid = self.worker.request_uid.next
        if self.is_single_multicall:
            return await self[0].get_response()
        
        if DEMO_MODE:
            # When demo mode is disabled, we can save some CPU time by skipping this sum
            demo_logger.info(f'request {rid} for jsonrpc batch {self.jid} ({sum(len(batch) for batch in self.calls)} calls) starting')  # type: ignore
        try:
            # NOTE: We do this inline so we don't have to allocate the response to memory
            await self.set_response(self.validate_responses(await self.post_sync()))
        except Exception as e:
            await (self.bisect_and_retry() if self.should_retry(e) else self.set_response(e))
        demo_logger.info(f'request {rid} for jsonrpc batch {self.jid} complete')  # type: ignore
    
    @eth_retry.auto_retry
    async def post_sync(self) -> Union[Dict, List[bytes]]:
        response = await self.worker.event_loop.run_in_executor(self.worker.executor, post_sync, self.controller.endpoint, self.data)
        if isinstance(response, (List, dict)):
            return response
        elif isinstance(response, Tuple):
            decoded, e = response
            counts = self.method_counts
            main_logger.error(f"{e.__class__.__name__}: {e}")
            main_logger.info(f"json batch id: {self.jid} | len: {len(self)} | total calls: {self.total_calls}", )
            main_logger.info(f"methods called: {counts}")
            if 'content length too large' in decoded or decoded == "":
                if self.is_multicalls_only:
                    self.controller.reduce_batch_size(self.total_calls)
                raise ValueError(decoded)
            # This shouldn't run unless there are issues. I'll probably delete it later.
            main_logger.info(f"decoded body: {decoded}")
            raise e
        raise NotImplementedError(response.__class__.__name__)
    
    @eth_retry.auto_retry
    async def post(self) -> Union[Dict, List[bytes]]:
        """ Posts `jsonrpc_batch` to your node. A successful call returns a list. """
        async with aiohttp.ClientSession(timeout=AIOHTTP_TIMEOUT) as session:
            responses = await session.post(self.controller.endpoint, json=self.data)  # type: ignore
            try:
                return await responses.json(content_type=responses.content_type)
            except Exception as e:
                counts = self.method_counts
                decoded = responses._body.decode()
                main_logger.info(f"json batch id: {self.jid} | len: {len(self)} | total calls: {self.total_calls}", )
                main_logger.info(f"methods called: {counts}")
                if 'content length too large' in decoded or decoded == "":
                    if self.is_multicalls_only:
                        self.controller.reduce_batch_size(self.total_calls)
                    raise ValueError(decoded) from e
                # This shouldn't run unless there are issues. I'll probably delete it later.
                main_logger.info(f"decoded body: {decoded}")
                main_logger.info(f"{e.__class__.__name__}: {e} {responses.content._exception}")
                raise 
    
    def should_retry(self, e: Exception) -> bool:
        # While it might look weird, f-string is faster than `str(e)`.
        if "No state available for block" in f"{e}":
            main_logger.debug('No state available for queried block. Bisecting batch and retrying.')
            return True
        elif f"{e}" == "jsonrpc":
            # TODO Figure out what this means and how we can prevent it.
            # For now, we simply bisect and retry.
            return True
        elif super().should_retry(e):
            return True
        return self.is_single_multicall
    
    async def set_response(self, response: Union[List[RPCResponse], Exception]) -> None:
        if isinstance(response, Exception):
            await await_all(call.set_response(response, wakeup_main_loop=False) for call in self.calls)
        else:
            await await_all(
                # NOTE: For some rpc methods, the result will be a dict we can't hash during the gather.
                call.set_response(AttributeDict(result["result"]) if isinstance(result["result"], dict) else result["result"], wakeup_main_loop=False)  # type: ignore
                for call, result in zip(self.calls, response)
            )
        self.set_done()
    
    def validate_responses(self, responses: T) -> T:
        # A successful response will be a list
        if isinstance(responses, dict) and 'result' in responses and isinstance(responses['result'], dict) and 'message' in responses['result']:
            raise ValueError(responses['result']['message'])
        for response in responses:
            if 'result' not in response:
                raise ValueError(response)
        return responses
    
    async def bisect_and_retry(self) -> None:
        await await_all(
            Multicall(self.worker, chunk[0].calls, f"json{self.jid}_{i}")  # type: ignore [misc]
            if len(chunk) == 1 and isinstance(chunk[0], Multicall)
            else JSONRPCBatch(self.worker, chunk, f"{self.jid}_{i}")
            for i, chunk in enumerate(self.bisect())
            if chunk
        )

    def _post_ensure_future_cleanup(self) -> None:
        self.controller.pending_rpc_calls = JSONRPCBatch(self.worker)
        


import queue
import threading
import weakref
from concurrent.futures import ThreadPoolExecutor, _base, thread

TEN_MINUTES = 60 * 10

def _worker(executor_reference, work_queue, initializer, initargs, timeout):  # NOTE: NEW 'timeout'
    if initializer is not None:
        try:
            initializer(*initargs)
        except BaseException:
            _base.LOGGER.critical('Exception in initializer:', exc_info=True)
            executor = executor_reference()
            if executor is not None:
                executor._initializer_failed()
            return
        
    try:
        while True:
            try:  # NOTE: NEW
                work_item = work_queue.get(block=True, 
                                           timeout=timeout)  # NOTE: NEW
            except queue.Empty:  # NOTE: NEW
                # Its been 'timeout' seconds and there are no new work items.  # NOTE: NEW
                executor = executor_reference()  # NOTE: NEW
                t = threading.current_thread()  # NOTE: NEW
                executor._threads.remove(t)  # NOTE: NEW
                thread._threads_queues.pop(t)  # NOTE: NEW
                return  # NOTE: NEW
                
            if work_item is not None:
                work_item.run()
                # Delete references to object. See issue16284
                del work_item

                # attempt to increment idle count
                executor = executor_reference()
                if executor is not None:
                    executor._idle_semaphore.release()
                del executor
                continue

            executor = executor_reference()
            # Exit if:
            #   - The interpreter is shutting down OR
            #   - The executor that owns the worker has been collected OR
            #   - The executor that owns the worker has been shutdown OR
            if thread._shutdown or executor is None or executor._shutdown:
                # Flag the executor as shutting down as early as possible if it
                # is not gc-ed yet.
                if executor is not None:
                    executor._shutdown = True
                # Notice other workers
                work_queue.put(None)
                return
            del executor
    except BaseException:
        _base.LOGGER.critical('Exception in worker', exc_info=True)

class PruningThreadPoolExecutor(ThreadPoolExecutor):
    """
    This ThreadPoolExecutor implementation prunes inactive threads after 'timeout' seconds without a work item.
    Pruned threads will be automatically recreated as needed for future workloads. up to 'max_threads' can be active at any one time.
    """
    def __init__(self, max_workers=None, thread_name_prefix='',
                 initializer=None, initargs=(), timeout=TEN_MINUTES):
        self._timeout=timeout
        super().__init__(max_workers, thread_name_prefix, initializer, initargs)
    
    def __repr__(self) -> str:
        return f"<dank_mids.{self.__class__.__name__} object at {hex(id(self))} [{len(self)} threads]>"
    
    def __len__(self) -> int:
        return len(self._threads)
        
    def _adjust_thread_count(self):
        # if idle threads are available, don't spin new threads
        if self._idle_semaphore.acquire(timeout=0):
            return

        # When the executor gets lost, the weakref callback will wake up
        # the worker threads.
        def weakref_cb(_, q=self._work_queue):
            q.put(None)

        num_threads = len(self._threads)
        if num_threads < self._max_workers:
            thread_name = '%s_%d' % (self._thread_name_prefix or self,
                                     num_threads)
            t = threading.Thread(name=thread_name, target=_worker,
                                 args=(weakref.ref(self, weakref_cb),
                                       self._work_queue,
                                       self._initializer,
                                       self._initargs,
                                       self._timeout))
            t.daemon = True
            t.start()
            self._threads.add(t)
            thread._threads_queues[t] = self._work_queue

executor = PruningThreadPoolExecutor(128)




from functools import lru_cache
from brownie.convert.utils import get_type_strings

from eth_abi import abi, codec


@lru_cache(maxsize=None)
def get_encoder(abi_inputs) -> codec.TupleEncoder:
    return codec.TupleEncoder(
        encoders=[abi.default_codec._registry.get_encoder(type_str) for type_str in get_type_strings(abi_inputs)]
    )

@lru_cache(maxsize=None)
def get_decoder(abi_inputs) -> codec.TupleDecoder:
    return codec.TupleDecoder(
        decoders=[abi.default_codec._registry.get_decoder(type_str) for type_str in get_type_strings(abi_inputs)]
    )


import asyncio as _asyncio
from typing import Awaitable as _Awaitable
from typing import Iterable as _Iterable

from tqdm.asyncio import tqdm_asyncio as _tqdm_asyncio

from dank_mids.helpers.w3 import setup_dank_w3, setup_dank_w3_from_sync


async def await_all(futs: _Iterable[_Awaitable], verbose: bool = False) -> None:
    # NOTE: 'verbose' is mainly for debugging but feel free to have fun
    if not isinstance(verbose, bool):
        raise NotImplementedError(verbose)
    as_completed = _tqdm_asyncio.as_completed if verbose else _asyncio.as_completed
    for fut in as_completed(futs if isinstance(futs, list) else [*futs]):
        await fut



from functools import lru_cache

from requests import Session
from requests.adapters import HTTPAdapter


@lru_cache(maxsize=1)
def get_session() -> Session:
    adapter = HTTPAdapter(pool_connections=100, pool_maxsize=100)
    session = Session()
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


import functools
from types import MethodType
from typing import Any, Coroutine, Dict, Optional, Tuple, Union

from brownie import Contract
from brownie.network.contract import ContractCall, ContractTx, OverloadedMethod
from dank_mids.brownie_patch.call import _patch_call
from web3 import Web3


def _patch_overloaded_method(call: OverloadedMethod, w3: Web3) -> None:
    @functools.wraps(call)
    async def coroutine(
        self: Contract,
        *args: Tuple[Any,...],
        block_identifier: Optional[Union[int, str, bytes]] = None,
        override: Optional[Dict[str, str]] = None
    ) -> Any:
        try:
            fn = self._get_fn_from_args(args)
        except ValueError as e:
            if f"Contract has more than one function '.{call._name}" in str(e) and f"You must explicitly declare which function you are calling, e.g. .{call._name}" in str(e):
                e = str(e)                    
                breakpoint = e.find("(*args)")
                raise ValueError(f"{e[:breakpoint]}.coroutine{e[breakpoint:]}")
            raise e
                                        
        kwargs = {"block_identifier": block_identifier, "override": override}
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return await fn.coroutine(*args, **kwargs)

    for args, method in call.__dict__['methods'].items():
        if isinstance(method, ContractCall) or isinstance(method, ContractTx):
            _patch_call(method, w3)
    # TODO implement this properly
        #elif isinstance(call, ContractTx):
            #_patch_tx(call, w3)
    
    call.coroutine = MethodType(coroutine, call)




from functools import lru_cache
from types import MethodType
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import eth_abi
from brownie.convert.normalize import ReturnValue
from brownie.exceptions import VirtualMachineError
from brownie.network.contract import ContractCall
from brownie.project.compiler.solidity import SOLIDITY_ERROR_CODES
from eth_abi import abi
from hexbytes import HexBytes
from multicall.utils import run_in_subprocess
from web3 import Web3
from web3.datastructures import AttributeDict

from dank_mids._config import BROWNIE_CALL_SEMAPHORE_VAL
from dank_mids.brownie_patch import _codec, _formatters
from dank_mids.semaphore import ThreadsafeSemaphore

stream = abi.default_codec.stream_class
brownie_call_semaphore = ThreadsafeSemaphore(BROWNIE_CALL_SEMAPHORE_VAL)

@lru_cache(maxsize=None)
def get_encode_fn(
    abi_name: str, 
    abi_inputs: Tuple[str, ...], 
    signature: str
) -> Callable[[Tuple[Any, ...]], str]:
    
    # NOTE: We define a function for each set of arg types so we don't need to parse the abi every call.
    if len(abi_inputs) == 0:
        async def encode_inputs(self: ContractCall, *args: Any):
            if len(args):
                raise TypeError(f"{abi_name} requires no arguments")
            return signature
    else:
        formatter = _formatters.get_abi_formatter(abi_inputs)
        encoder = _codec.get_encoder(abi_inputs)
        def __encode_inputs(*args) -> str:
            # Format contract inputs based on ABI types
            # NOTE: We've removed a bunch of if statements from the formatters to speed things up.
            try:
                return signature + encoder(formatter(args)).hex()
            except Exception as e:
                raise type(e)(f"{abi_name} {e}") from None
            
        async def encode_inputs(self: ContractCall, *args: Any) -> str:
            return await run_in_subprocess(__encode_inputs, *args)
            
    return encode_inputs


@lru_cache(maxsize=None)
def get_decode_fn(
    abi_outputs: Tuple[str, ...]
) -> Callable[[str], Any]:
    
    # NOTE: We define a function for each set of arg types so we don't need to parse the abi every call.
    decoder = _codec.get_decoder(abi_outputs)
    formatter = _formatters.get_abi_formatter(abi_outputs)
    
    def __decode_output(hexstr: str) -> Any:
        selector = HexBytes(hexstr)[:4].hex()
        if selector == "0x08c379a0":
            revert_str = eth_abi.decode_abi(["string"], HexBytes(hexstr)[4:])[0]
            raise ValueError(f"Call reverted: {revert_str}")
        elif selector == "0x4e487b71":
            error_code = int(HexBytes(hexstr)[4:].hex(), 16)
            if error_code in SOLIDITY_ERROR_CODES:
                revert_str = SOLIDITY_ERROR_CODES[error_code]
            else:
                revert_str = f"Panic (error code: {error_code})"
            raise ValueError(f"Call reverted: {revert_str}")
        if abi_outputs and not hexstr:
            raise ValueError("No data was returned - the call likely reverted")
        
        result = ReturnValue(formatter(decoder(stream(HexBytes(hexstr)))), abi_outputs)
        if len(result) == 1:
            result = result[0]
        return result
    
    async def decode_output(self: ContractCall, hexstr: str) -> Any:
        return await run_in_subprocess(__decode_output, hexstr)
    
    return decode_output
    

@lru_cache(maxsize=None)
def get_coroutine_fn(w3: Web3):
    assert w3.eth.is_async, "You must provide a w3 instance with an async provider."
    async def coroutine(
        self: ContractCall,
        *args: Tuple[Any,...],
        block_identifier: Optional[Union[int, str, bytes]] = None,
        override: Optional[Dict[str, str]] = None
    ) -> Any:
        if override:
            raise ValueError("Cannot use state override with `coroutine`.")
        
        async with brownie_call_semaphore:
            try:
                return await self._decode_output(
                    await w3.eth.call({"to": self._address, "data": await self._encode_input(*args)}, block_identifier)  # type: ignore
                )
            except ValueError as e:
                try:
                    raise VirtualMachineError(e) from None
                except:
                    raise e
    return coroutine

def _make_abi_hashable(
    abi: List[Dict[str, str]]
) -> Tuple[AttributeDict, ...]:
    """Takes a list of dicts and returns a hashable tuple of AttributeDicts."""
    return tuple(
        AttributeDict({k: _make_abi_hashable(v) if isinstance(v, list) else v for k, v in x.items()})
        for x in abi
    )

def _patch_call(call: ContractCall, w3: Web3) -> None:
    """Patch a brownie 'ContractCall' object with a 'coroutine' method for async calling."""
    # NOTE: We define abi-specific encode and decode functions in the hope that it speeds up execution.
    #       Haven't tested the impact but it definitely removes some boilerplate code from script execution which should only mean good things.
    encode_fn = get_encode_fn(call.abi['name'], _make_abi_hashable(call.abi['inputs']), call.signature)
    decode_fn = get_decode_fn(_make_abi_hashable(call.abi['outputs']))
    call._encode_input = MethodType(encode_fn, call)
    call._decode_output = MethodType(decode_fn, call)
    call.coroutine = MethodType(get_coroutine_fn(w3), call)





from typing import Any, Callable

from eth_utils.curried import (apply_formatter_if, apply_formatters_to_dict,
                               apply_key_map, is_null)
from eth_utils.toolz import complement, compose
from hexbytes import HexBytes
from web3 import Web3
from web3._utils.rpc_abi import RPC
from web3.types import RPCEndpoint

from dank_mids.helpers._middleware.formatting import \
    async_construct_formatting_middleware
from dank_mids.types import AsyncMiddleware

is_not_null = complement(is_null)

remap_geth_poa_fields = apply_key_map(
    {
        "extraData": "proofOfAuthorityData",
    }
)

pythonic_geth_poa = apply_formatters_to_dict(
    {
        "proofOfAuthorityData": HexBytes,
    }
)

geth_poa_cleanup = compose(pythonic_geth_poa, remap_geth_poa_fields)

async def geth_poa_middleware(make_request: Callable[[RPCEndpoint, Any], Any], w3: Web3) -> AsyncMiddleware:
    middleware = await async_construct_formatting_middleware(
        result_formatters={
            RPC.eth_getBlockByHash: apply_formatter_if(is_not_null, geth_poa_cleanup),
            RPC.eth_getBlockByNumber: apply_formatter_if(is_not_null, geth_poa_cleanup),
        },
    )
    return await middleware(make_request, w3)  # type: ignore [arg-type, return-value]


import asyncio
import threading
from typing import TYPE_CHECKING, Any, Coroutine, Generator, List

import eth_retry
from eth_typing import ChecksumAddress
from multicall.multicall import NotSoBrightBatcher
from web3 import Web3

from dank_mids import _config, executor
from dank_mids.helpers import await_all
from dank_mids.loggers import main_logger
from dank_mids.requests import JSONRPCBatch, Multicall, _Batch
from dank_mids.types import Multicalls
from dank_mids.uid import UIDGenerator

if TYPE_CHECKING:
    from dank_mids.controller import DankMiddlewareController


def call(endpoint: str, *request_args):
    return Web3(endpoint).eth.call(*request_args)


class DankWorker:
    """
    Runs a second event loop in a subthread which is used to reduce congestion on the main event loop.
    A second loop with less items running in parallel allows dank_mids to get your responses back to your
      main loop faster.
    """
    def __init__(self, controller: "DankMiddlewareController") -> None:
        self.controller = controller
        self.target: ChecksumAddress = self.controller.multicall2
        self.batcher = NotSoBrightBatcher()
        self.multicall_uid: UIDGenerator = UIDGenerator()
        self.request_uid: UIDGenerator = UIDGenerator()
        self.jsonrpc_batch_uid: UIDGenerator = UIDGenerator()
        self.state_override_supported: bool = not _config.GANACHE_FORK and self.controller.chain_id != 100  # Gnosis Chain does not support state override.
        self.event_loop = asyncio.new_event_loop()
        self.worker_thread = threading.Thread(target=self.start)
        self.worker_thread.start()
        self.executor = executor.executor
            
    def start(self) -> None:
        """ Runs in worker thread. """
        asyncio.set_event_loop(self.event_loop)
        self.event_loop.run_until_complete(self.loop())
    
    async def loop(self) -> None:
        """ Exits loop when main thread dies, killing worker thread. Runs in worker thread. """
        while threading.main_thread().is_alive():
            self.controller._wakeup_loop()
            await asyncio.sleep(5)
    
    @eth_retry.auto_retry
    async def __call__(self, *request_args: Any) -> Any:
        # NOTE: We make the actual reuest synchronously so we can get the results from the node without waiting for the event loop
        return await self.run_in_executor(self.controller.sync_w3.eth.call, *request_args)  # type: ignore

    async def run_in_executor(self, fn, *args, loop=None):
        return await (loop or self.event_loop).run_in_executor(self.executor, fn, *args)
    
    def run_coro_threadsafe(self, coro: Coroutine):
        # NOTE: This function should be called in the main thread only. Coro will execute in worker thread.
        fut = asyncio.run_coroutine_threadsafe(coro, loop=self.event_loop)
        fut.add_done_callback(self._raise_exception_in_main_thread)
        return fut
    
    async def execute_batch(self, calls_to_exec: Multicalls, rpc_calls: JSONRPCBatch) -> None:
        """ Runs in main thread. """
        # NOTE: We materialize the generator here so that the generator execution
        #      takes place in the main thread leaving the worker less congested.
        coros = [*DankBatch(self, calls_to_exec, rpc_calls).coroutines]
        self.controller._futs.append(self.run_coro_threadsafe(await_all(coros)))
        
    def _raise_exception_in_main_thread(self, fut: asyncio.Future) -> None:
        """Callback used to raise any exceptions that occur in the worker thread in the main thread."""
        if e := fut.exception():
            main_logger.error(e)
            raise e
        try:
            self.controller._futs.remove(fut)
        except ValueError as e:
            if str(e) != "list.remove(x): x not in list":
                raise
            
class DankBatch:
    """ A batch of jsonrpc batches. """
    def __init__(self, worker: DankWorker, eth_calls: Multicalls, rpc_calls: JSONRPCBatch):
        self.worker = worker
        for mcall in eth_calls.values():
            for call in mcall:
                call._started = True
        for call in rpc_calls:
            call._started = True
        self.eth_calls = eth_calls
        self.rpc_calls = rpc_calls
    
    #def __await__(self) -> Generator[Any, None, Any]:
    #    return await_all(self.coroutines).__await__()
    
    @property
    def batcher(self) -> NotSoBrightBatcher:
        return self.worker.batcher
    
    @property
    def coroutines(self) -> Generator["_Batch", None, None]:
        *full_batches, working_batch = self.batch_multicalls(list(self.eth_calls.values()))
        
        # Yield full batches then yield the rest
        yield from full_batches
        if len(working_batch) + len(self.rpc_calls) <= _config.MAX_JSONRPC_BATCH_SIZE:
            working_batch.extend(self.rpc_calls, skip_check=True)
            yield working_batch
        else:
            yield working_batch
            yield self.rpc_calls
    
    def batch_multicalls(self, multicalls: List["Multicall"]) -> Generator["JSONRPCBatch", None, None]:
        """ Used to collect multicalls into batches without overwhelming the node with oversized calls. """
        multicalls.sort(key=lambda x: len(x), reverse=True)
        eth_calls_in_batch = 0
        working_batch = JSONRPCBatch(self.worker)
        for mcall in multicalls:
            assert isinstance(mcall, Multicall)
            # This would be too many eth_calls for a single multicall, lets start a new jsonrpc batch.
            if eth_calls_in_batch + len(mcall) > self.batcher.step:
                if working_batch:
                    yield working_batch
                working_batch = JSONRPCBatch(self.worker, [mcall])
                eth_calls_in_batch = len(mcall)
            # This can be added to the current jsonrpc batch
            else:
                working_batch.append(mcall)
                eth_calls_in_batch += len(mcall)
                if len(working_batch) >= _config.MAX_JSONRPC_BATCH_SIZE:
                    # There are more than `MAX_JSONRPC_BATCH_SIZE` rpc calls packed into this batch, let's start a new one
                    yield working_batch
                    working_batch = JSONRPCBatch(self.worker)
                    eth_calls_in_batch = 0
        yield working_batch




from typing import Optional, Union

from brownie import Contract, network, web3
from brownie.network.contract import ContractCall, ContractTx, OverloadedMethod
from dank_mids.brownie_patch.call import _patch_call
from dank_mids.brownie_patch.overloaded import _patch_overloaded_method
from web3 import Web3

ContractMethod = Union[ContractCall, ContractTx, OverloadedMethod]

def _patch_if_method(method: ContractMethod, w3: Web3) -> None:
    if isinstance(method, ContractCall) or isinstance(method, ContractTx):
        _patch_call(method, w3)
    elif isinstance(method, OverloadedMethod):
        _patch_overloaded_method(method, w3)


def patch_contract(contract: Contract, w3: Optional[Web3]) -> Contract:
    if w3 is None and network.is_connected():
        w3 = web3
    assert w3 is not None, "You must make sure either brownie is connected or you pass in a Web3 instance."
    for k, v in contract.__dict__.items():
        _patch_if_method(v, w3)
    return contract
